# -*- coding: utf-8 -*-

import os
import sys
import inspect
import cPickle

#
#
class PluginMgr:
    '''
    プラグイン管理マネージャ。
    
    生成時に指定されたプラグインディレクトリ中から、プラグインプロトコル
    に適合するプラグインモジュールをロードする。
    ロードしたプラグインモジュールに関して、モジュール名、モジュール中の
    関数名と関数オブジェクトを辞書として保持する。

    各モジュールのどの関数をどのタイミングで使用するかはこのクラスを利用する
    アプリケーションごとのプラグインプロトコル次第で決まるため、このクラスでは
    ケアしない。

    また、プラグイン設定保存・読込み用の関数名は、プラグインプロトコルに依存するため、
    プラグイン設定の保存・読込みは自動では行われない。このクラスを使用する側が必要に
    応じて、保存・読み込み用メソッドの呼出す必要がある。
    '''
    def __init__(self, targetDir, pluginFuncs):
        '''
        初期化。

        targetDir:
            管理対象のプラグインディレクトリパス。
        requiredFuncs:
            プラグインプロトコルに適合するための関数情報のリスト。
            [(関数名, 必須フラグ), ....]
            必須フラグがTrueに指定されている関数はプロトコル適合のためには必須。
            Falseのものは、その関数が存在すれば呼び出されることを示す。
        '''

        # プラグインプロトコルチェック用リスト
        # 必須関数用リスト
        self.requiredFuncNames = []

        # その他の関数用リスト
        self.otherFuncNames = []

        # プロトコル関数情報の設定
        self.set_plugin_funcs(pluginFuncs)
                        
        # 管理ディレクトリパスを初期化
        self.pluginDirPath = None

        # プラグイン情報用の辞書を初期化
        # {'モジュール名': {関数名: 関数オブジェクト, ...},
        #  'モジュール名': {関数名: 関数オブジェクト, ...}}
        # という管理構造となっている。
        self.pluginDict = {}

        # プラグインディレクトリの登録
        self.set_plugin_dir(targetDir)

        # パスが実存するディレクトリの場合には、その中に存在するプラグイン
        # を走査し、自身の管理用辞書に登録する
        self.load_plugins()

    def set_plugin_funcs(self, funcs):
        '''
        プラグインプロトコルに適合する関数条件を設定する。

        funcs:
            プラグインプロトコルに適合するための関数情報のリスト。
            [(関数名, 必須フラグ), ....]
        Return:
            None
        '''        
        if funcs == None:
            return            

        # 現状の情報を破棄
        del self.requiredFuncNames[:]
        del self.otherFuncNames[:]

        # 必須関数名、その他の関数名を振り分けておく
        for name, flag in funcs:
            if flag:
                self.requiredFuncNames.append(name)
            else:
                self.otherFuncNames.append(name)

    def set_plugin_dir(self, path):
        '''
        パス文字列を正規化し、プラグインディレクトリとして登録。

        path:
            プラグインディレクトリのパス文字列。
        Return:
            None
        '''
        if path == None or len(path) == 0:
            return

        # プラグインディレクトリパスを正規化
        self.pluginDirPath = os.path.normpath(os.path.expanduser(path))

        # 管理対象のプラグインディレクトリがモジュールパスに含まれて
        # いない場合には追加する。
        if self.pluginDirPath not in sys.path:
            sys.path.append(self.pluginDirPath)

    def load_plugins(self):
        '''
        プラグインディレクトリ中のプラグインをロードし、
        プラグインのプロトコルチェックを通過したもののみを
        自身の管理用辞書に追加する。

        Return:
            None
        '''
        if self.pluginDirPath == None:
            return
        if os.path.isdir(self.pluginDirPath) == False:
            return

        # プラグイン管理用辞書をクリア
        self.pluginDict.clear()

        # プラグインディレクトリ中のファイルを調べ、プラグインプロトコルに
        # 適合するモジュールの情報を構築する。
        for item in os.listdir(self.pluginDirPath):
            if os.path.isfile(item) != False:
                continue

            modName = inspect.getmodulename(item)
            
            funcInfoDict = self.import_plugin(modName)
            if funcInfoDict:
                self.pluginDict[modName] = funcInfoDict

    def import_plugin(self, modName):
        '''
        プラグインプロトコルに一致したクラスをモジュールからインポートする。

        modName:
            モジュール名文字列。
        Return:
            辞書。
            {関数名: 関数オブジェクト, ....}
            モジュール名が不正な場合やプラグインプロトコルに適合しなかった
            場合にはNoneが返される。
        '''
        if modName == None or len(modName) == 0:
            return None

        # モジュールをインポート
        try:
            mod = __import__(modName)
        except:
            print 'error on importing %s' % modName
            return None

        # モジュールからクラス属性をもつオブジェクトを取得する。
        funcList = inspect.getmembers(mod, inspect.isfunction)
        if funcList == None or len(funcList) == 0:
            return None

        # 必須の関数が存在しない場合にはそのモジュールは
        # プロトコルに対応していないと判定し、処理を中断。
        # (そのモジュールの関数情報は作成しない)
        funcNames = [item[0] for item in funcList]
        for fname in self.requiredFuncNames:
            if fname not in funcNames:
                return None

        # モジュール内の関数のうち、プロトコル適合条件として
        # 登録されている関数についてのみ情報を登録する。
        # (プロトコルに無関係な関数オブジェクトは無視する)
        funcDict = {}
        for name, funcObj in funcList:
            if (name not in self.requiredFuncNames and
                name not in self.otherFuncNames):
                continue

            funcDict[name] = funcObj

        return funcDict

    def get_plugin_count(self):
        '''
        登録されているプラグイン数を取得する。

        Return:
            整数。プラグインモジュール数。
        '''
        return len(self.pluginDict)
    
    def get_plugin_names(self):
        '''
        登録されているプラグイン名のリストを取得する。

        Return:
            リスト。[モジュール名, ...]
        '''
        return self.pluginDict.keys()

    def get_plugin_func_dict(self, modName):
        '''
        プラグインモジュール名（文字列）を指定して、そのモジュール
        のプラグイン関数情報を取得する。

        modName:
            モジュール名文字列
        return:
            辞書
            {関数名: 関数オブジェクト, ....}
        '''
        if modName == None:
            return None
        if modName not in self.pluginDict:
            return None

        return self.pluginDict[modName]

    def get_plugin_func(self, modName, funcName):
        '''
        プラグインモジュール名、関数名を指定して、モジュール中の
        関数を取得する。

        modName:
            モジュール名文字列
        funcName:
            関数名文字列
        return:
            関数
            指定された関数が存在しない場合にはNoneを返す
        '''
        if modName == None or funcName == None:
            return None

        funcDict = self.get_plugin_func_dict(modName)
        if funcDict == None:
            return None

        if funcName not in funcDict:
            return None
        
        return funcDict[funcName]
    
    def check_config_args(self, confDir, modName, funcName):
        '''
        プラグイン設定・取得関数の引数をチェックする。
        '''
        if (confDir == None or
            modName == None or
            funcName == None):
            print 'illegal param'
            return False

        if os.path.isdir(confDir) == False:
            return False

        return True

    def load_all_configs(self, confDir, funcName):
        '''
        管理している全てのプラグインの設定を読み込む。
        
        confDir:
            設定ファイル出力先ディレクトリ
        funcName:
            プラグインの現在の設定を取得するための関数名
        return:
            なし
        '''
        for modName in self.get_plugin_names():
            self.load_config(confDir, modName, funcName)
    
    def load_config(self, confDir, modName, funcName):
        '''
        modNameで指定されたプラグインの設定(pickle化されて保存されている)を
        読み込み、プラグインに設定する。
        プラグインへの設定はfuncNameで指定された名前の関数を使用する。
        この関数はconfDir中に"モジュール名.pkl"が存在することを期待する。
        存在しない場合は、読み込みをスキップする。

        confDir:
            設定ファイル読み込みディレクトリ
        modName:
            設定を読み込むモジュール名
        funcName:
            プラグインの現在の設定を設定するための関数名
        return:
            True  成功
            False 失敗
        '''
        # 引数チェック
        if self.check_config_args(confDir, modName, funcName) == False:
            return False

        # プラグインの関数辞書を取得し、モジュールに指定された関数が
        # 存在するかチェック。
        func_load = self.get_plugin_func(modName, funcName)
        if func_load == None:
            print '%s does not have such function(%s).' % (modName, funcName)
            return False
            
        # 設定ファイルパスを作成して、設定ファイルの存在確認。
        # 設定ファイルが存在しない場合には処理を中断。
        fname = os.path.join(confDir, modName+'.pkl')
        if os.path.isfile(fname) == False:
            #print 'config file(%s) is not found, skipped.' % fname
            return False
        
        try:
            # pickle化されている設定ファイルをunpickeして、
            # Pluginに設定
            f = file(fname, 'rb')
            confData = cPickle.load(f)
            f.close()
            if confData:
                func_load(confData)
        except:
            print 'Can not load the plugin configuration.'
            return False

        return True

    def save_all_configs(self, confDir, funcName):
        '''
        管理中の全てのプラグイン設定を保存する。

        confDir:
            設定ファイル出力先ディレクトリ
        funcName:
            プラグインの現在の設定を取得するための関数名
        return:
            なし
        '''
        for modName in self.get_plugin_names():
            self.save_config(confDir, modName, funcName)
    
    def save_config(self, confDir, modName, funcName):
        '''        
        modNameで指定されたプラグイン設定をpickle化してファイルへ保存する。
        プラグインの設定は、funcNameで指定された名前の関数を使用して
        プラグインから取得する。
        保存される設定ファイルの名前は"モジュール名.pkl"となる。

        confDir:
            設定ファイル出力先ディレクトリ
        modName:
            設定を保存するモジュール名
        funcName:
            プラグインの現在の設定を取得するための関数名
        return:
            True  成功
            False 失敗
        '''
        # 引数チェック
        if self.check_config_args(confDir, modName, funcName) == False:
            return False

        # プラグインの関数辞書を取得し、モジュールに指定された関数が
        # 存在するかチェック。
        func_save = self.get_plugin_func(modName, funcName)
        if func_save == None:
            print '%s does not have such function(%s), skipped.' % (modName, funcName)
            return False

        # プラグインから設定データを取得する。
        # 設定データが空っぽの場合には、保存処理をスキップする。
        confData = func_save()
        if confData:
            try:
                f = file(os.path.join(confDir, modName+'.pkl'), 'wb')
                cPickle.dump(confData, f)
                f.close()
            except:
                print 'Can not save the plugin configuration.'
                return False
        else:
            #print 'Config is None, skipped.'
            pass
            
        return True

#
# 動作テスト
#
if __name__ == '__main__':
    def test_plugin_mgr_protocol(dirpath, funcs, confDir, getFunc, setFunc):
        # プラグインマネージャを生成
        mgr = PluginMgr(dirpath, funcs)

        #
        # 生成したプラグインマネージャの情報を確認
        #
        print '------------------------'
        print 'plugin module count : %d' % mgr.get_plugin_count()

        for name in mgr.get_plugin_names():
            print 'name: %s' % name
            funcDict = mgr.get_plugin_func_dict(name)
            for funcName, funcObj in funcDict.iteritems():
                print '\t%s(...): %s' % (funcName, funcObj)

            mgr.load_config(confDir, name, setFunc)
            mgr.save_config(confDir, name, getFunc)
            mgr.load_config(confDir, name, setFunc)


    # プラグインプロトコルに適合する関数の条件を準備
    funcs = [('get_name', True),
             ('get_version', True),
             ('get_widgets', True),
             ('get_author', False),
             ('get_description', False),
             ('get_settings', False),
             ('set_settings', False)]

    # 正常系
    test_plugin_mgr_protocol('./plugins', funcs,
                             './configs', 'get_settings', 'set_settings')

    # ディレクトリパスが不正
    test_plugin_mgr_protocol('./aaa', funcs,
                             './configs', 'get_settings', 'set_settings')

    # プラグインプロトコル関数リストが不正
    test_plugin_mgr_protocol('./plugins', None,
                             './configs', 'get_settings', 'set_settings')

    test_plugin_mgr_protocol('./plugins', [],
                             './configs', 'get_settings', 'set_settings')
